/*  vim: set sw=8 ts=8 si et :
* Author: Guido Socher, Copyright: GPL 2
* mini voltmeter with
* LED display.
* http://tuxgraphics.org/electronics/
*
* Chip type           : Atmega8
* Clock frequency     : Internal clock 8 Mhz 
*/
#include <avr/io.h>
#include <inttypes.h>
#include <stdlib.h>
#include "timeout.h"
#include "onewire.h"
#include "ds18x20.h"

//---
// Do no decimal point. Do not insert any jumpers.
//
// No decimal point:
#define DP_DIG 0
//
#define MAXSENSORS 3 // just the array size
static uint8_t gSensorIDs[MAXSENSORS][OW_ROMCODE_SIZE];
static int8_t gNsensors=0;
static int16_t gTempdata=0; // temperature


//Connections of the seven segment elements: 
//      10 9 7 6 4 2 1 = pin
//   0| -  f a b c d e | 0111111 | 3f
//   1| -  - - b c - - | 0001100 | 0c
//   2| g  - a b - d e | 1011011 | 5b
//   3| g  - a b c d - | 1011110 | 5e
//   4| g  f - b c - - | 1101100 | 6c
//   5| g  f a - c d - | 1110110 | 76
//   6| g  f a - c d e | 1110111 | 77
//   7| -  - a b c - - | 0011100 | 1c
//   8| g  f a b c d e | 1111111 | 7f
//   9| g  f a b c d - | 1111110 | 7e
//   -| g  - - - - - - | 1000000 | 40 minus=0x40
// digit to seven segment LED mapping:
static unsigned char d2led[]={0x3f,0x0c,0x5b,0x5e,0x6c,0x76,0x77,0x1c,0x7f,0x7e};

void delay_ms(unsigned char ms)
{
        // if we would use the library function _delay_ms
        // everywhere then the code would become very fat
        // We just need integers therefore we wrap it.
        while(ms){
                _delay_ms(0.96);
                ms--;
        }
}

void init7segment(void)
{
        // least significant digit, cathode transistor:
        DDRD|= (1<<DDD7);
        // middle digit:
        DDRB|= (1<<DDB6);
        // left digit:
        DDRB|= (1<<DDB7);
        // digit pins:
        DDRD|= 0x7F;
        // off
        PORTD=0;
}

// display: --
void outofrange(void){
        PORTD=0x40; // just a dash
        PORTD|= (1<<PD7);// on
        delay_ms(4);
        PORTD&=~(1<<PD7);
        PORTB|= (1<<PB6);// on for a short time
        delay_ms(4);
        PORTB &= ~(1<<PB6);
}

// Update the sevent segement display.
// You must loop over this function all the time.
inline void upd7segment(int16_t number)
{
        uint8_t minus=0;
        if (number<0){
                number=abs(number);
                if (number>99){
                        // out of range for 3 digits
                        number=0;
                }
                minus=1;
        }
        // LS digit:
        PORTD=d2led[(number%10)];
        //
        PORTD|=(1<<PD7);// digit on for a short time:
        delay_ms(4);
        number/=10; // divide by 10 (shift off right most digit)
        PORTD&=~(1<<PD7);
#if (DP_DIG == 0)
        if (number>0){ // chop off the twp left most digits if it is zero
#endif
                // middle digit:
                PORTD=d2led[(number%10)];
                //
                PORTB|= (1<<PB6);// digit on for a short time:
                delay_ms(4);
                number/=10; // divide by 10 (shift off right most digit)
                PORTB &= ~(1<<PB6);
                // minus on the left
                if (minus){
                        PORTD=0x40;
                        PORTB|= (1<<PB7);// on
                        delay_ms(4);
                        PORTB &= ~(1<<PB7);
                        // go out of this function
                        return;
                }
#if (DP_DIG == 1 || DP_DIG == 0)
                if (number>0){ // chop off left most digit if it is zero
#endif
                        //
                        // left digit, we assume there that the number is
                        // smaller than 999, no modulo here:
                        PORTD=d2led[number];
                        //
                        PORTB|= (1<<PB7);// digit on for a short time:
                        delay_ms(4);
                        PORTB &= ~(1<<PB7);
#if (DP_DIG == 1 || DP_DIG == 0)
                }
#endif
#if (DP_DIG == 0)
        }else{
                if (minus){
                        PORTD=0x40;
                        PORTB|= (1<<PB6);// on
                        delay_ms(4);
                        PORTB &= ~(1<<PB6);
                }
        }
#endif
}

// writes to global array gSensorIDs
int8_t search_sensors(void)
{
	uint8_t diff;
        uint8_t nSensors=0;
	for( diff = OW_SEARCH_FIRST; 
		diff != OW_LAST_DEVICE && nSensors < MAXSENSORS ; )
	{
		DS18X20_find_sensor( &diff, &(gSensorIDs[nSensors][0]) );
		
		if( diff == OW_PRESENCE_ERR ) {
			return(-1); //No Sensor found
		}
		if( diff == OW_DATA_ERR ) {
			return(-2); //Bus Error
		}
		nSensors++;
	}
	return nSensors;
}

// read the latest measurement off the scratchpad of the ds18x20 sensor
// and store it in gTempdata
void read_temp_meas(void){
        uint8_t subzero, cel, cel_frac_bits;
        // we have exactly 1 sensor
        if ( DS18X20_read_meas( &gSensorIDs[0][0], &subzero,
                        &cel, &cel_frac_bits) == DS18X20_OK ) {
                gTempdata=cel;
                // round to next degree if above 0.5'C, cel_frac_bits has
                // 16 steps.
                if (cel_frac_bits>7){
                        gTempdata++;
                }
                if (subzero){
                        gTempdata=-gTempdata;
                }
        }else{
                gTempdata=0;
        }
}

// convert celsius values to Farenheit
int16_t c2f(int16_t celsius){
        return((celsius * 18)/10) + 32;
}

/*  we display the temperature reading from a ds18x20 sensor.
 *  The Sensor connected as follows:
 *  Sensor GND to GND
 *  Sensor data to PB2
 *  Sensor Vcc to PB1
 *
 * Pin PC1 can be used to switch between 'C and 'F. By default
 * the display is in 'C.
 *
 */
int main(void)
{
        unsigned int lc; // loop counter
        unsigned char sensorfail=1; 
        // PB2 = ds18x20 sensor data pin, input with pullup
        DDRB&= ~(1<<PINB2);
        PORTB|=1<<PINB2; // internal pullup resistor on
        // PB1 = output = +5 for sensor
        DDRB|= (1<<DDB1);
        PORTB|=(1<<PORTB1);
        // Pc1 = input with pullup, switch C' to F'
        DDRC&= ~(1<<PINC1);
        PORTC|=1<<PINC1; // internal pullup resistor on
        //
        gNsensors=search_sensors();
        init7segment();
        if (gNsensors==1){
                if ( DS18X20_start_meas(NULL) == DS18X20_OK) {
                        sensorfail=0;
                }
        }
        // we have to wait 750ms or more before reading the result:
        lc=80;
        while (lc) {
                lc--;
                outofrange(); // about 10ms
        }
        lc=0;
        while (1) {
                // upd7segment gives between 7 to 20ms delay
                if (lc==0){
                        // read previous measurement, it should be ready 
                        // now as minimum 750ms have already passed.
                        // result goes to gTempdata
                        read_temp_meas();
                        if (bit_is_clear(PINC,PINC1)){
                                gTempdata=c2f(gTempdata);
                        }
                }
                if (lc==1){
                        // start new measurement.
                        // we can only detect a short circuit here. No sensor
                        // shows up as zero.
                        if ( DS18X20_start_meas(NULL) == DS18X20_OK) {
                                sensorfail=0;
                        }else{
                                sensorfail=1;
                        }
                }
                if (sensorfail==0){
                        upd7segment(gTempdata);
                }else{
                        outofrange();
                }
                lc++;
                if (lc>500){
                        lc=0;
                }
        }
        return(0);
}

